/*
   myFBO.cpp
   openFrameworks

    Simple OpenGL FBO Ping-Pong Class
    Copyright (C) 2008  Matthias Dörfelt, www.mokafolio.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


 */

#include "myFBO.h"


void myFBO::init(int w, int h, bool dbuffer){

// currently this class needs non Power of two texture support -> if its not available it wont run

	if (!GLEE_ARB_texture_non_power_of_two)
    {
        printf("GL_EXT_framebuffer_object extension is unavailable!\n");
        exit(0);
    }

    if (!GLEE_EXT_framebuffer_object)
    {
        printf("GL_EXT_framebuffer_object extension is unavailable!\n");
        exit(0);
	}

	clean();

	width = w;
	height = h;

	read_tex = 0;
	write_tex = 1;

	dbuffer = dbuffer;

	// Setup our FBO
	glGenFramebuffersEXT(1, &fbo);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);

	// setup a two textures for ping-pong rendering
	glGenTextures(2, (GLuint *)textureName);

	//setup two color buffer render targets
	att_point[0] = GL_COLOR_ATTACHMENT0_EXT;
	att_point[1] = GL_COLOR_ATTACHMENT1_EXT;


	glEnable(GL_TEXTURE_2D);

	// init the two textures
	glBindTexture(GL_TEXTURE_2D, (GLuint)textureName[0]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, 0);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glBindTexture(GL_TEXTURE_2D, (GLuint)textureName[1]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, 0);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glGenerateMipmapEXT(GL_TEXTURE_2D);

	// attach the textures to the FBO
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, att_point[read_tex], GL_TEXTURE_2D, (GLuint)textureName[0], 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, att_point[write_tex], GL_TEXTURE_2D, (GLuint)textureName[1], 0);



	// attach depthrenderbuffer to our fbo if needed
	if(dbuffer){
	glGenRenderbuffersEXT(1, &depthBuffer);
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depthBuffer);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, width, height);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, depthBuffer);
	}

	checkStatus();

	clear();

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);	// Unbind the FBO for now

	glDisable(GL_TEXTURE_2D);
}

void myFBO::checkStatus(){

// FBO check method from http://www.openframeworks.cc/forum/viewtopic.php?t=371&start=0
GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);

   switch(status)
   {
      case GL_FRAMEBUFFER_COMPLETE_EXT:
         break;

      case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
         printf("framebuffer GL_FRAMEBUFFER_UNSUPPORTED_EXT\n");
         /* you gotta choose different formats */
         break;

      case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
         printf("framebuffer INCOMPLETE_ATTACHMENT\n");
         break;

      case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
         printf("framebuffer FRAMEBUFFER_MISSING_ATTACHMENT\n");
         break;

      case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
         printf("framebuffer FRAMEBUFFER_DIMENSIONS\n");
         break;

      case 0x8CD8: //GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT:
         printf("framebuffer INCOMPLETE_DUPLICATE_ATTACHMENT\n");
         break;

      case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
         printf("framebuffer INCOMPLETE_FORMATS\n");
         break;

      case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
         printf("framebuffer INCOMPLETE_DRAW_BUFFER\n");
         break;

      case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
         printf("framebuffer INCOMPLETE_READ_BUFFER\n");
         break;

      case GL_FRAMEBUFFER_BINDING_EXT:
         printf("framebuffer BINDING_EXT\n");
         break;

      case 0x8CDE: // GL_FRAMEBUFFER_STATUS_ERROR_EXT:
         printf("framebuffer STATUS_ERROR\n");
         break;

      default:
         /* programming error; will fail on all hardware */
         //exit(0);
         break;
    }
}

void myFBO::swapIn()
{
    // bind FBO
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);

    // set the camera settings according to the default ones of OpenFrameworks
	glViewport( 0, 0, width, height);

	float halfFov, theTan, screenFov, aspect;
	screenFov 		= 60.0f;

	float eyeX 		= (float)width / 2.0;
	float eyeY 		= (float)height / 2.0;
	halfFov 		= PI * screenFov / 360.0;
	theTan 			= tanf(halfFov);
	float dist 		= eyeY / theTan;
	float nearDist 	= dist / 10.0;
	float farDist 	= dist * 10.0;
	aspect 			= (float)width/(float)height;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(screenFov, aspect, nearDist, farDist);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(eyeX, eyeY, dist,
	eyeX, eyeY, 0.0, 0.0, 1.0, 0.0);


	glScalef(1, -1, 1);
  	glTranslatef(0, -height, 0);

	//select texture to draw to.
	glDrawBuffer(att_point[write_tex]);
}

void myFBO::swapOut()
{

	//unbind FBO
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	//restore default view
	glViewport( 0, 0, ofGetWidth(), ofGetHeight() );
	ofSetupScreen();

}

void myFBO::clean()
{
	if (textureName[0] != 0 || textureName[1] != 0){
		glDeleteTextures(2, (GLuint *)textureName);
	}
	width = 0;
	height = 0;
}

void myFBO::clear()
{
//	glClearColor(0,0,0,0); //clear color (bg of the FBO tex)
//	glClear(GL_COLOR_BUFFER_BIT);	// Clear Screen data of the texture to write on
//	if(dbuffer) glClear(GL_DEPTH_BUFFER_BIT); //Clear Depth Buffer if available

	glClearColor(1,1,1,0); //clear white
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	// Clear Screen And Depth Buffer
}

void myFBO::render(float x, float y, float w, float h, bool flipY){
   float tcy1, tcy2, tcy3, tcy4;
   if(flipY){
   tcy1 = 0.0;
   tcy2 = 0.0;
   tcy3 = 1.0;
   tcy4 = 1.0;
   } else {
   tcy1 = 1.0;
   tcy2 = 1.0;
   tcy3 = 0.0;
   tcy4 = 0.0;
   }
   glEnable(GL_TEXTURE_2D);
   glActiveTexture(GL_TEXTURE0);
   glBindTexture(GL_TEXTURE_2D, (GLuint)textureName[write_tex]);
   glBegin(GL_QUADS);
			glTexCoord2f(0.0,tcy1);			glVertex3f(x, y,0);
			glTexCoord2f(1.0,tcy2);			glVertex3f(x+w, y,0);
			glTexCoord2f(1.0,tcy3);			glVertex3f(x+w, y+h,0);
			glTexCoord2f(0.0,tcy4);			glVertex3f(x, y+h,0);
   glEnd();
   glBindTexture(GL_TEXTURE_2D, 0);
   glDisable(GL_TEXTURE_2D);

}

void myFBO::swapTex(){

//swap texture targets
if(write_tex == 0){
write_tex = 1;
read_tex = 0;
} else {
write_tex = 0;
read_tex = 1;
}
}


